import os
import random
from hashlib import sha1

M = 3
N = 64
n = 2

class Permutation:
	def __init__(self, L):
		self.n = len(L)
		self.L = L
		assert all(i in L for i in range(self.n))

	def __mul__(self, other):
		assert self.n == other.n
		return Permutation([other.L[self.L[i]] for i in range(self.n)])

	def __eq__(self, other):
		return self.L == other.L

	def inv(self):
		return Permutation([self.L.index(i) for i in range(self.n)])

	def cycles(self):
		elts = list(range(self.n))
		cycles = []
		while len(elts) > 0:
			cur = []
			i = elts[0]
			while i not in cur:
				cur.append(i)
				elts.remove(i)
				i = self.L[i]
			cycles.append(cur)
		# cycles.sort(key=len)
		return cycles

	def __getitem__(self, i):
		return self.L[i]

	def __str__(self):
		return "".join("({})".format(" ".join(str(e) for e in c)) for c in self.cycles())

	def __repr__(self):
		return "Permutation({})".format(self.L)


def random_permutation(n):
	random.seed(os.urandom(100))
	L = list(range(n))
	for i in range(n-1):
		j = random.randint(i, n-1)
		L[i], L[j] = L[j], L[i]
	return Permutation(L)

# for i in range(100):
#     x = random_permutation(N)
#     assert x * x.inv() == Permutation(list(range(N)))

def encrypt(m, k):
	x, y = m
	for i in range(M):
		x, y = (y, x * k.inv() * y * k)
	return x, y

def decrypt(c, k):
	x, y = c
	for i in range(M):
		x, y = (y * k.inv() * x.inv() * k, x)
	return x, y

# if __name__ == "__main__":
#     k = random_permutation(N)
#     print "The flag is: PCTF{%s}" % sha1(str(k)).hexdigest()
#     pairs = []
#     for i in range(n):
#         pt = random_permutation(N), random_permutation(N)
#         ct = encrypt(pt, k)
#         assert pt == decrypt(ct, k)
#         pairs.append((pt,ct))

#     with open("data.txt", "w") as f:
#         f.write(str(pairs))

def conjugate(x, k):
	return k.inv() * x * k


def rotate(l):
	return l.append(l.pop(0))

def get_conjugate(lhs, rhs):
	d = {}
	for c1, c2 in zip(lhs, rhs):
		for e1, e2 in zip(c1, c2):
			d[e1] = e2
	k = Permutation([d[i] for i in range(N)])
	return k


def clock(l, r, i):
	if i >= len(l):
		return
	rotate(l[i])
	r[i] += 1
	if r[i] == len(l[i]):
		r[i] = 0
		clock(l, r, i+1)

def final_solve(lhs, l, test_lhs, test_rhs):
	r = [0 for i in range(len(l))]
	while r[-1] != len(l[-1])-1:
		go(lhs, l, test_lhs, test_rhs)
		clock(l, r, 0)


def go(lhs, rhs, test_lhs, test_rhs):
	k = get_conjugate(lhs, rhs)
	if test_rhs == conjugate(test_lhs, k):
		print "The flag is: PCTF{%s}" % sha1(str(k)).hexdigest()
		print



p1, c1 = ((Permutation([48, 27, 39, 5, 49, 32, 26, 23, 19, 22, 28, 63, 60, 18, 35, 59, 15, 52, 11, 0, 12, 50, 46, 13, 25, 47, 14, 54, 42, 16, 29, 37, 31, 4, 21, 61, 40, 6, 30, 62, 2, 10, 45, 56, 3, 36, 17, 8, 55, 20, 7, 43, 9, 51, 53, 58, 44, 34, 57, 1, 33, 41, 38, 24]), Permutation([28, 34, 61, 25, 57, 56, 59, 7, 6, 27, 62, 0, 54, 10, 36, 23, 21, 38, 35, 40, 30, 45, 60, 55, 22, 5, 52, 29, 11, 17, 44, 31, 63, 42, 41, 51, 20, 3, 13, 14, 46, 37, 2, 48, 32, 26, 15, 33, 19, 49, 50, 4, 8, 47, 43, 16, 1, 53, 39, 12, 24, 18, 58, 9])), (Permutation([5, 13, 36, 51, 3, 63, 59, 53, 45, 52, 37, 10, 39, 15, 41, 16, 57, 49, 34, 21, 40, 2, 44, 55, 6, 24, 43, 23, 48, 25, 8, 60, 26, 62, 19, 12, 33, 7, 29, 30, 27, 18, 0, 46, 61, 11, 50, 4, 32, 38, 54, 17, 14, 35, 42, 9, 47, 20, 28, 31, 56, 58, 1, 22]), Permutation([25, 45, 37, 52, 27, 26, 9, 55, 40, 15, 59, 30, 3, 13, 62, 8, 17, 53, 47, 56, 6, 20, 11, 0, 24, 21, 39, 33, 19, 32, 41, 44, 43, 5, 18, 60, 58, 12, 23, 42, 46, 22, 36, 16, 28, 63, 38, 10, 49, 4, 14, 1, 34, 57, 7, 54, 48, 29, 2, 31, 35, 61, 51, 50])))

x, y = p1
x3, y3 = c1

lhs = y * x3
rhs = x.inv() * y3

print lhs
print rhs
print
# print get_conjugate(lhs.cycles(), rhs.cycles())


print
p2, c2 = ((Permutation([27, 11, 22, 17, 49, 33, 48, 12, 14, 6, 62, 53, 41, 5, 24, 13, 21, 46, 36, 61, 29, 60, 58, 43, 16, 47, 45, 20, 39, 37, 19, 31, 10, 42, 44, 54, 51, 15, 0, 34, 35, 18, 8, 26, 30, 23, 3, 40, 32, 52, 28, 1, 7, 2, 25, 59, 38, 9, 56, 50, 4, 57, 63, 55]), Permutation([50, 62, 22, 26, 33, 21, 35, 23, 53, 45, 47, 24, 41, 10, 38, 31, 2, 15, 55, 32, 34, 1, 25, 49, 52, 6, 37, 58, 7, 19, 29, 17, 16, 3, 4, 11, 63, 8, 27, 0, 59, 18, 54, 12, 61, 39, 13, 28, 20, 51, 36, 42, 9, 56, 46, 60, 30, 5, 48, 57, 43, 40, 14, 44])), (Permutation([63, 61, 38, 9, 30, 51, 39, 33, 45, 24, 0, 5, 17, 6, 23, 20, 49, 22, 18, 29, 37, 59, 28, 31, 36, 26, 2, 13, 53, 52, 34, 19, 25, 44, 16, 27, 43, 55, 8, 42, 12, 21, 62, 46, 7, 4, 56, 35, 41, 10, 3, 11, 57, 50, 15, 14, 48, 40, 60, 1, 47, 32, 54, 58]), Permutation([44, 31, 39, 23, 24, 26, 32, 57, 35, 55, 17, 3, 42, 7, 33, 14, 30, 47, 21, 56, 50, 62, 58, 11, 60, 36, 37, 63, 20, 2, 15, 25, 5, 34, 0, 22, 28, 45, 40, 38, 46, 41, 9, 59, 1, 8, 27, 49, 52, 4, 6, 16, 10, 48, 43, 29, 12, 13, 54, 51, 18, 19, 53, 61])))

x, y = p2
x3, y3 = c2
lhs2 = y * x3
rhs2 = x.inv() * y3

print lhs2
print rhs2
print
# print get_conjugate(lhs2.cycles(), rhs2.cycles())

print rhs2.cycles()
rhs2 = [[58], [5, 7, 10], [11, 31, 25, 43], [2, 48, 32, 52, 4, 18, 41, 42, 34, 38, 12, 57, 19, 15, 45, 37],[1, 16, 60, 62, 17, 23, 8, 9, 13, 14, 35, 46, 47, 36, 21, 30], [0, 40, 49, 24, 33, 26, 59, 29, 50, 51, 28, 6, 55, 61, 56, 54, 22, 39, 20, 63, 53, 3, 27, 44]]

lhs2 = lhs2.cycles()
lhs2.sort(key=len)
print lhs2
final_solve(lhs2, rhs2, lhs, rhs)